import {Injectable} from '@angular/core';
import {Product} from "../model/product";
import {Job, JobSpec} from "../model/job";
import {BehaviorSubject, Observable} from "rxjs";
import {StorageService} from "./storage.service";
import {Epoch} from "../model/epoch";
import {Tile} from "../model/tile";


@Injectable({
  providedIn: 'root'
})
export class FiltersService {

  private filters = {
    'EPOCH': Epoch.EPOCH,
    'PRODUCT_TYPE': Product.TYPES,
    'PRODUCT_STATUS': ['ALL', 'COMPLETED', 'WAITING', 'READY', 'COMPLETED', 'PROCESSING', 'CACHED'],
    'JOB_QUEUE': Job.QUEUES,
    'JOB_STATUS': ['ALL', 'WAITING', 'PROCESSING', 'QA_READY', 'QA_MANUAL', 'QA_ACCEPTED', 'QA_REJECTED', 'QA_ON_HOLD', 'QA_MANUAL_ACCEPTED', 'ERROR'],
    'SORT_DIRECTION': ['ASC', 'DESC'],
    'TILE_SORT': Tile.SORT_COLUMNS,
    'PRODUCT_SORT': Product.SORT_COLUMNS,
    'JOBSPEC_SORT': JobSpec.SORT_COLUMNS,
    'JOB_SORT': Job.SORT_COLUMNS
  };

  private defaultSettings = {
    'EPOCH': Epoch.getEpochFromId(2),
    'PRODUCT_TYPE': Product.getTypeFromName('calibration'),
    'PRODUCT_STATUS': 'ALL',
    'JOB_QUEUE': Job.getQueueFromName('calibration'),
    'JOB_STATUS': 'ALL',
    'SORT_DIRECTION': 'ASC',
    'TILE_SORT': 'id',
    'PRODUCT_SORT': 'id',
    'JOBSPEC_SORT': 'id',
    'JOB_SORT': 'id'
  };

  private currentSettings = {};
  private _currentSettings: BehaviorSubject<object> = new BehaviorSubject<object>(this.currentSettings);
  public readonly currentSettings$: Observable<object> = this._currentSettings.asObservable();

  constructor(private storageService: StorageService) {
    // fill the currentSettings with default values
    this.currentSettings = this.defaultSettings;

    // see if we have any saved preferences
    let savedFilters = this.storageService.getLocal('filters');
    if (savedFilters) {
      savedFilters = JSON.parse(savedFilters);
      for (let sf in savedFilters) {
        if (this.filters.hasOwnProperty(sf)) {
          // now loop through the allowed values for this setting
          for (let allowed of this.filters[sf]) {
            // if we have a match, it's safe to set
            // Because objects here are statically defined, I feel it's safe to use JSON.stringify to compare
            if (JSON.stringify(allowed) === JSON.stringify(savedFilters[sf])) {
              this.currentSettings[sf] = allowed;
            }
          }
        }
      }
    }
    this._currentSettings.next(this.currentSettings);
  }

  static prettyName(name: string): string {
    if (name.length < 3) {
      return name.toUpperCase();
    } else {
      name = name.replace(/_/g, ' ');
      return name.toLowerCase().split(' ').map(word => word.charAt(0).toUpperCase() + word.slice(1)).join(' ');
    }
  }

  getFilters() {
    return this.filters
  }

  getFilter(key: string): Array<any> {
    key = key.toUpperCase();
    // ensure this is a valid setting
    if (this.filters.hasOwnProperty(key)) {
      return this.filters[key];
    }
    return null;
  }

  getFilterKeys(): Array<string> {
    return Object.keys(this.filters);
  }

  getCurrentSettings(): object {
    return this.currentSettings;
  }

  getCurrentSetting(key: string): any {
    key = key.toUpperCase();
    // ensure this is a valid setting
    if (this.filters.hasOwnProperty(key)) {
      // if we don't have a value, set it with the default
      if (!this.currentSettings.hasOwnProperty(key)) {
        this.currentSettings[key] = this.defaultSettings[key];
      }
      return this.currentSettings[key];
    }
    return null;
  }

  setCurrentSetting(key: any, value: any): boolean {
    key = key.toUpperCase();
    // make sure this is a valid setting
    if (this.filters.hasOwnProperty(key)) {
      // now loop through the allowed values for this setting
      for (let allowed of this.filters[key]) {
        // if we have a match, it's safe to set
        if ((typeof allowed === 'string' && typeof value === 'string' && allowed.toLowerCase() === value.toLowerCase())
          || allowed === value) {
          this.currentSettings[key] = allowed;
          // There is somewhat of a link between PRODUCT_TYPE and JOB_QUEUE
          if (key === 'PRODUCT_TYPE') {
            const jobQueue = Job.getQueueFromName(allowed.name);
            if (jobQueue) {
              this.currentSettings['JOB_QUEUE'] = jobQueue;
            }
          } else if (key === 'JOB_QUEUE') {
            const productType = Product.getTypeFromName(allowed.name);
            if (productType) {
              this.currentSettings['PRODUCT_TYPE'] = productType;
            }
          }
          this.storageService.saveLocal('filters', JSON.stringify(this.currentSettings));
          this._currentSettings.next(this.currentSettings);
          return true;
        }
      }
    }
    return false;
  }

}
